@@ -1,7 +1,4 @@
-try:
-    import ujson as json
-except ImportError:
-    import json
+import ujson
 import urllib
 import os
 import httplib
@@ -12,7 +9,9 @@ import urllib2
 import jsonpatch
 from boto.s3.connection import S3Connection, OrdinaryCallingFormat
 import boto
-import filechunkio
+from cStringIO import StringIO
+
+from . import __version__
 
 
 
@@ -36,12 +35,14 @@ class Item(object):
 
     # init()
     #_____________________________________________________________________________________
-    def __init__(self, identifier):
+    def __init__(self, identifier, metadata_timeout=None):
         self.identifier = identifier
         self.details_url = 'https://archive.org/details/{0}'.format(identifier)
         self.download_url = 'https://archive.org/download/{0}'.format(identifier)
         self.metadata_url = 'https://archive.org/metadata/{0}'.format(identifier)
+        self.metadata_timeout = metadata_timeout
         self._s3_conn = None
+        self._bucket = None
         self.metadata = self._get_item_metadata()
         if self.metadata == {}:
             self.exists = False
@@ -52,8 +53,8 @@ class Item(object):
     # _get_item_metadata()
     #_____________________________________________________________________________________
     def _get_item_metadata(self):
-        f = urllib.urlopen(self.metadata_url)
-        return json.loads(f.read())
+        f = urllib2.urlopen(self.metadata_url, timeout=self.metadata_timeout)
+        return ujson.loads(f.read())
 
 
     # files()
@@ -132,7 +133,7 @@ class Item(object):
             patch.append(dict((k,v) for k,v in pd.items() if v))
 
         data = {
-            '-patch': json.dumps(patch),
+            '-patch': ujson.dumps(patch),
             '-target': target,
             'access': access_key,
             'secret': secret_key,
@@ -153,7 +154,7 @@ class Item(object):
         self.metadata = self._get_item_metadata()
         return dict(
             status_code = status_code,
-            content = json.loads(resp_file.read()),
+            content = ujson.loads(resp_file.read()),
         )
 
 
@@ -171,44 +172,28 @@ class Item(object):
     def _get_s3_bucket(self, conn, headers={}, ignore_bucket=False):
         if ignore_bucket is True:
             headers['x-archive-ignore-preexisting-bucket'] = 1
-            bucket = None
+            self._bucket = None
         else:
-            bucket = conn.lookup(self.identifier)
-        if bucket:
-            return bucket
-        bucket = conn.create_bucket(self.identifier, headers=headers)
+            if self._bucket is None:
+                self._bucket = conn.lookup(self.identifier)
+        if self._bucket:
+            return self._bucket
+        self._bucket = conn.create_bucket(self.identifier, headers=headers)
         i=0
         while i<60:
             b = conn.lookup(self.identifier)
             if b:
-                return bucket
+                return self._bucket
             time.sleep(10)
             i+=1
         raise NameError('Could not create or lookup {0}'.format(self.identifier))
 
 
-    # upload()
+    # _get_s3_headers()
     #_____________________________________________________________________________________
-    def upload(self, files, meta_dict={}, headers={}, dry_run=False,
-               derive=True, multipart=False, ignore_bucket=False):
-        """Upload file(s) to an item. The item will be created if it does not
-        exist.
-
-        :param files: Either a list of filepaths, or a string pointing to a single file.
-        :param meta_dict: Dictionary of metadata used to create a new item.
-        :param dry_run: (optional) Boolean. Set to True to print headers to stdout -- don't upload anything.
-        :param derive: (optional) Boolean. Set to False to prevent an item from being derived after upload.
-        :param multipart: (optional) Boolean. Set to True to upload files in parts. Useful when uploading large files.
-        :param ignore_bucket: (optional) Boolean. Set to True to ignore and clobber existing files and metadata.
-
-        Usage::
-
-            >>> import internetarchive
-            >>> item = internetarchive.Item('identifier')
-            >>> item.upload('/path/to/image.jpg', dict(mediatype='image', creator='Jake Johnson'))
-        """
-        # Convert metadata from :meta_dict: into S3 headers
-        for key,v in meta_dict.iteritems():
+    def _get_s3_headers(self, headers, metadata):
+        """Convert metadata from :metadata: into S3 headers"""
+        for key,v in metadata.iteritems():
             if type(v) == list:
                 for i, value in enumerate(v):
                     s3_header_key = 'x-archive-meta{0:02d}-{1}'.format(i, key)
@@ -222,51 +207,110 @@ class Item(object):
                     headers[s3_header_key] = v.encode('utf-8')
                 else:
                     headers[s3_header_key] = v
-        headers = dict((k, v) for k, v in headers.iteritems() if v)
+        return dict((k, v) for k, v in headers.iteritems() if v)
 
-        if dry_run:
-            return headers
 
+    # upload_file()
+    #_____________________________________________________________________________________
+    def upload_file(self, _file, remote_name=None, metadata={}, headers={}, derive=True,
+                    ignore_bucket=False, multipart=False, bytes_per_chunk=16777216,
+                    debug=False):
+        """Upload a single file to an item. The item will be created if it does not exist.
+
+        :param _file: String or File. The filepath or file-like object to be uploaded.
+        :param remote_name: (optional) String. Sets the remote filename.
+        :param metadata: (optional) Dictionary. Metadata used to create a new item.
+        :param headers: (optional) Dictionary. Add additional IA-S3 headers to request.
+        :param derive: (optional) Boolean. Set to False to prevent an item from being derived after upload.
+        :param multipart: (optional) Boolean. Set to True to upload files in parts. Useful when uploading large files.
+        :param ignore_bucket: (optional) Boolean. Set to True to ignore and clobber existing files and metadata.
+        :param debug: (optional) Boolean. Set to True to print headers to stdout -- don't upload anything.
+        :param bytes_per_chunk: (optional) Integer. Used to determine the chunk size when using multipart upload.
+
+        Usage::
+
+            >>> import internetarchive
+            >>> item = internetarchive.Item('identifier')
+            >>> item.upload_file('/path/to/image.jpg', remote_name='photos/image1.jpg')
+            True
+
+        """
+
+        headers = self._get_s3_headers(headers, metadata)
+        headers['x-archive-size-hint'] = os.stat(_file).st_size
+        scanner = 'Internet Archive Python library {0}'.format(__version__)
+        headers['x-archive-meta-scanner'] = scanner
+
+        if type(_file) == str:
+            _file = file(_file, 'rb')
+        if remote_name is None:
+            remote_name = _file.name.split('/')[-1]
+
+        conn = self._get_s3_conn()
+        bucket = self._get_s3_bucket(conn, headers, ignore_bucket=ignore_bucket)
+
+        if derive is False:
+            headers['x-archive-queue-derive'] =  0
+
+        # Don't clobber existing files unless ignore_bucket is True.
+        if bucket.get_key(remote_name) and ignore_bucket is False:
+            return True
+
+        if multipart is False:
+            k = boto.s3.key.Key(bucket)
+            k.name = remote_name
+            k.set_contents_from_file(_file, headers=headers)
+        else:
+            mp = bucket.initiate_multipart_upload(remote_name, headers=headers)
+            def read_chunk():
+                return _file.read(bytes_per_chunk)
+            part = 1
+            for chunk in iter(read_chunk, ''):
+                part_fp = StringIO(chunk)
+                mp.upload_part_from_file(part_fp, part_num=part)
+                part += 1
+            mp.complete_upload()
+            #    mp.cancel_upload()
+        return True
+
+
+    # upload()
+    #_____________________________________________________________________________________
+    def upload(self, files, metadata={}, headers={}, derive=True, ignore_bucket=False, 
+               multipart=False, bytes_per_chunk=16777216, debug=False):
+        """Upload files to an item. The item will be created if it does not exist.
+
+        :param files: List. The filepaths or file-like objects to upload.
+        :param metadata: (optional) Dictionary. Metadata used to create a new item.
+        :param derive: (optional) Boolean. Set to False to prevent an item from being derived after upload.
+        :param ignore_bucket: (optional) Boolean. Set to True to ignore and clobber existing files and metadata.
+        :param multipart: (optional) Boolean. Set to True to upload files in parts. Useful when uploading large files.
+        :param bytes_per_chunk: (optional) Integer. Used to determine the chunk size when using multipart upload.
+        :param debug: (optional) Boolean. Set to True to print headers to stdout -- don't upload anything.
+
+        Usage::
+
+            >>> import internetarchive
+            >>> item = internetarchive.Item('identifier')
+            >>> md = dict(mediatype='image', creator='Jake Johnson')
+            >>> item.upload('/path/to/image.jpg', md, derive=False)
+            True
+
+        """
+
+        if debug is True:
+            return self._get_s3_headers(headers, metadata)
         if type(files) != list:
             files = [files]
-        for f in files:
-            if type(f) == str:
-                f = file(f, 'rb')
-            filename = f.name.split('/')[-1]
-            conn = self._get_s3_conn()
-            bucket = self._get_s3_bucket(conn, headers, ignore_bucket=ignore_bucket)
-
-            if not derive:
-                headers = {'x-archive-queue-derive': 0}
-
-            if bucket.get_key(filename) and ignore_bucket is False:
+        for _file in files:
+            upload_status = self.upload_file(_file, remote_name=None, metadata=metadata, 
+                                                    headers=headers, derive=derive, 
+                                                    ignore_bucket=ignore_bucket,
+                                                    multipart=multipart) 
+            if upload_status is True:
                 continue
-
-            if not multipart:
-                k = boto.s3.key.Key(bucket)
-                k.name = filename
-                k.set_contents_from_file(f, headers=headers)
             else:
-                mp = bucket.initiate_multipart_upload(filename, headers=headers)
-                source_size = os.stat(f).st_size
-                headers['x-archive-size-hint'] = source_size
-                bytes_per_chunk = (max(int(math.sqrt(5242880) *
-                                           math.sqrt(source_size)), 5242880))
-                chunk_amount = (int(math.ceil(source_size/float(bytes_per_chunk))))
-                for i in range(chunk_amount):
-                    offset = i * bytes_per_chunk
-                    remaining_bytes = source_size - offset
-                    part_bytes = min([bytes_per_chunk, remaining_bytes])
-                    part_num = i + 1
-                    with filechunkio.FileChunkIO(f, 'r',
-                                                 offset=offset,
-                                                 bytes=part_bytes) as f:
-                        mp.upload_part_from_file(fp=f, part_num=part_num)
-                if len(mp.get_all_parts()) == chunk_amount:
-                    mp.complete_upload()
-                    key = bucket.get_key(filename)
-                else:
-                    mp.cancel_upload()
+                return False
         return True
 
 
@@ -285,7 +329,7 @@ class File(object):
         self.size = file_dict.get('size')
         if self.size is not None:
             self.size = int(self.size)
-        self.format = file_dict.get('format') 
+        self.format = file_dict.get('format')
         self.mtime = file_dict.get('mtime')
         self.md5  = file_dict.get('md5')
         self.sha1 = file_dict.get('crc32')
@@ -317,7 +361,7 @@ class Search(object):
     for archive.org items using the advanced search engine:
         >>> import internetarchive
         >>> search = internetarchive.Search('(uploader:jake@archive.org)')
-        >>> for result in search.results: 
+        >>> for result in search.results:
         ...     print result['identifier']
     """
 
@@ -349,11 +393,11 @@ class Search(object):
         info_params['rows'] = 0
         encoded_info_params = urllib.urlencode(info_params)
         f = urllib.urlopen(self._base_url, encoded_info_params)
-        results = json.loads(f.read())
+        results = ujson.loads(f.read())
         del results['response']['docs']
         return results
 
-        
+
     # _iter_results()
     #_____________________________________________________________________________________
     def _iter_results(self):
@@ -363,7 +407,7 @@ class Search(object):
             self.params['page'] = page
             encoded_params = urllib.urlencode(self.params)
             f = urllib.urlopen(self._base_url, encoded_params)
-            results = json.loads(f.read())
+            results = ujson.loads(f.read())
             for doc in results['response']['docs']:
                 yield doc
 
@@ -371,17 +415,17 @@ class Search(object):
 # Catalog class
 #_________________________________________________________________________________________
 class Catalog(object):
+    GREEN = 0
+    BLUE = 1
+    RED = 2
+    BROWN = 9
 
     # init()
     #_____________________________________________________________________________________
     def __init__(self, params=None):
         url = 'http://archive.org/catalog.php'
-        self.GREEN = 0
-        self.BLUE = 1
-        self.RED = 2
-        self.BROWN = 9
 
-        if not params:
+        if params is None:
             params = dict(justme = 1)
 
         # Add params required to retrieve JSONP from the IA catalog.
@@ -402,20 +446,63 @@ class Catalog(object):
         jsonp_str = f.read()
         json_str = jsonp_str[(jsonp_str.index("(") + 1):jsonp_str.rindex(")")]
 
-        tasks_json = json.loads(json_str)
-        self.tasks = [dict(
-            identifier = t[0],
-            server = t[1],
-            command = t[2],
-            time = t[3],
-            submitter = t[4],
-            # Parse args into dict
-            args = dict(x for x in urllib2.urlparse.parse_qsl(t[5])),
-            task_id = t[6],
-            row_type = t[7],
-        ) for t in tasks_json]
+        tasks_json = ujson.loads(json_str)
+        self.tasks = [CatalogTask(t) for t in tasks_json]
+        
+    def filter_tasks(self, pred):
+        return [t for t in self.tasks if pred(t)]
 
-        self.green_rows = [t for t in self.tasks if t['row_type'] == self.GREEN]
-        self.blue_rows = [t for t in self.tasks if t['row_type'] == self.BLUE]
-        self.red_rows = [t for t in self.tasks if t['row_type'] == self.RED]
-        self.brown_rows = [t for t in self.tasks if t['row_type'] == self.BROWN]
+    def tasks_by_type(self, row_type):
+        return self.filter_tasks(lambda t: t.row_type == row_type)
+
+    @property
+    def green_rows(self):
+        return self.tasks_by_type(self.GREEN)
+    @property
+    def blue_rows(self):
+        return self.tasks_by_type(self.BLUE)
+    @property
+    def red_rows(self):
+        return self.tasks_by_type(self.RED)
+    @property
+    def brown_rows(self):
+        return self.tasks_by_type(self.BROWN)
+
+class CatalogTask(object):
+    """represents catalog task.
+    """
+    COLUMNS = ('identifier', 'server', 'command', 'time', 'submitter',
+               'args', 'task_id', 'row_type')
+
+    def __init__(self, columns):
+        """:param columns: array of values, typically returned by catalog
+        web service. see COLUMNS for the column name.
+        """
+        for a, v in map(None, self.COLUMNS, columns):
+            if a: setattr(self, a, v)
+        # special handling for 'args' - parse it into a dict if it is a string
+        if isinstance(self.args, basestring):
+            self.args = dict(x for x in urllib2.urlparse.parse_qsl(self.args))
+
+    def __repr__(self):
+        return ('CatalogTask(identifier={identifier},'
+                ' task_id={task_id!r}, server={server!r},'
+                ' command={command!r},'
+                ' submitter={submitter!r},'
+                ' row_type={row_type})'.format(**self.__dict__))
+
+    def __getitem__(self, k):
+        """dict-like access privided as backward compatibility."""
+        if k in self.COLUMNS:
+            return getattr(self, k, None)
+        else:
+            raise KeyError, k
+
+    def open_task_log(self):
+        """return file-like reading task log."""
+        if self.task_id is None:
+            raise ValueError, 'task_id is None'
+        url = 'http://catalogd.archive.org/log/{0}'.format(self.task_id)
+        return urllib2.urlopen(url)
+
+        
